@@ -10,7 +10,6 @@
 -----------------------------------------------------------------------------
 
 {-# LANGUAGE ScopedTypeVariables #-}
-{-# LANGUAGE DeriveDataTypeable  #-}
 {-# LANGUAGE DefaultSignatures   #-}
 
 module Data.SBV.SMT.SMT where
@@ -22,20 +21,23 @@ import Control.DeepSeq    (NFData(..))
 import Control.Monad      (when, zipWithM)
 import Data.Char          (isSpace)
 import Data.Int           (Int8, Int16, Int32, Int64)
-import Data.List          (intercalate, isPrefixOf, isInfixOf)
+import Data.Function      (on)
+import Data.List          (intercalate, isPrefixOf, isInfixOf, sortBy)
 import Data.Word          (Word8, Word16, Word32, Word64)
 import System.Directory   (findExecutable)
-import System.Process     (runInteractiveProcess, waitForProcess, terminateProcess)
+import System.Environment (getEnv)
 import System.Exit        (ExitCode(..))
 import System.IO          (hClose, hFlush, hPutStr, hGetContents, hGetLine)
+import System.Process     (runInteractiveProcess, waitForProcess, terminateProcess)
 
 import qualified Data.Map as M
-import Data.Typeable
 
 import Data.SBV.BitVectors.AlgReals
 import Data.SBV.BitVectors.Data
 import Data.SBV.BitVectors.PrettyNum
-import Data.SBV.Utils.Lib (joinArgs)
+import Data.SBV.BitVectors.Symbolic   (SMTEngine)
+import Data.SBV.SMT.SMTLib            (interpretSolverOutput, interpretSolverModelLine)
+import Data.SBV.Utils.Lib             (joinArgs, splitArgs)
 import Data.SBV.Utils.TDiff
 
 -- | Extract the final configuration from a result
@@ -53,6 +55,9 @@ newtype ThmResult    = ThmResult    SMTResult
 -- The reason for having a separate 'SatResult' is to have a more meaningful 'Show' instance.
 newtype SatResult    = SatResult    SMTResult
 
+-- | A 'safe' call results in a 'SafeResult'
+newtype SafeResult   = SafeResult   (Maybe String, String, SMTResult)
+
 -- | An 'allSat' call results in a 'AllSatResult'. The boolean says whether
 -- we should warn the user about prefix-existentials.
 newtype AllSatResult = AllSatResult (Bool, [SMTResult])
@@ -69,6 +74,14 @@ instance Show SatResult where
                                      "Unknown"     "Unknown. Potential model:\n"
                                      "Satisfiable" "Satisfiable. Model:\n" r
 
+-- | User friendly way of printing safety results
+instance Show SafeResult where
+   show (SafeResult (mbLoc, msg, r)) = showSMTResult (tag "No violations detected")
+                                                     (tag "Unknown")  (tag "Unknown. Potential violating model:\n")
+                                                     (tag "Violated") (tag "Violated. Model:\n") r
+        where loc   = maybe "" (++ ": ") mbLoc
+              tag s = loc ++ msg ++ ": " ++ s
+
 -- | The Show instance of AllSatResults. Note that we have to be careful in being lazy enough
 -- as the typical use case is to pull results out as they become available.
 instance Show AllSatResult where
@@ -89,22 +102,6 @@ instance Show AllSatResult where
                            Satisfiable{} -> True
                            _             -> False
 
--- | The result of an 'sAssert' call
-data SafeResult = SafeNeverFails
-                | SafeAlwaysFails  String
-                | SafeFailsInModel String SMTConfig SMTModel
-                deriving Typeable
-
--- | The show instance for SafeResult. Note that this is for display purposes only,
--- user programs are likely to pattern match on the output and proceed accordingly.
-instance Show SafeResult where
-   show SafeNeverFails              = "No safety violations detected."
-   show (SafeAlwaysFails s)         = intercalate "\n" ["Assertion failure: " ++ show s, "*** Fails in all assignments to inputs"]
-   show (SafeFailsInModel s cfg md) = intercalate "\n" ["Assertion failure: " ++ show s, showModel cfg md]
-
--- | If a 'prove' or 'sat' call comes accross an 'sAssert' call that fails, they will throw a 'SafeResult' as an exception.
-instance C.Exception SafeResult
-
 -- | Instances of 'SatModel' can be automatically extracted from models returned by the
 -- solvers. The idea is that the sbv infrastructure provides a stream of 'CW''s (constant-words)
 -- coming from the solver, and the type @a@ is interpreted based on these constants. Many typical
@@ -308,8 +305,8 @@ instance Modelable SMTResult where
   getModel (ProofError _ s)  = error $ unlines $ "Backend solver complains: " : s
   getModel (TimeOut _)       = Left "Timeout"
   getModel (Satisfiable _ m) = Right (False, parseModelOut m)
-  modelExists (Satisfiable{}) = True
-  modelExists (Unknown{})     = False -- don't risk it
+  modelExists Satisfiable{}   = True
+  modelExists Unknown{}       = False -- don't risk it
   modelExists _               = False
   getModelDictionary (Unsatisfiable _) = M.empty
   getModelDictionary (Unknown _ m)     = M.fromList (modelAssocs m)
@@ -337,23 +334,35 @@ displayModels disp (AllSatResult (_, ms)) = do
 -- | Show an SMTResult; generic version
 showSMTResult :: String -> String -> String -> String -> String -> SMTResult -> String
 showSMTResult unsatMsg unkMsg unkMsgModel satMsg satMsgModel result = case result of
-  Unsatisfiable _                   -> unsatMsg
-  Satisfiable _ (SMTModel [] [] []) -> satMsg
-  Satisfiable _ m                   -> satMsgModel ++ showModel cfg m
-  Unknown _ (SMTModel [] [] [])     -> unkMsg
-  Unknown _ m                       -> unkMsgModel ++ showModel cfg m
-  ProofError _ []                   -> "*** An error occurred. No additional information available. Try running in verbose mode"
-  ProofError _ ls                   -> "*** An error occurred.\n" ++ intercalate "\n" (map ("***  " ++) ls)
-  TimeOut _                         -> "*** Timeout"
+  Unsatisfiable _             -> unsatMsg
+  Satisfiable _ (SMTModel []) -> satMsg
+  Satisfiable _ m             -> satMsgModel ++ showModel cfg m
+  Unknown     _ (SMTModel []) -> unkMsg
+  Unknown     _ m             -> unkMsgModel ++ showModel cfg m
+  ProofError  _ []            -> "*** An error occurred. No additional information available. Try running in verbose mode"
+  ProofError  _ ls            -> "*** An error occurred.\n" ++ intercalate "\n" (map ("***  " ++) ls)
+  TimeOut     _               -> "*** Timeout"
  where cfg = resultConfig result
 
--- | Show a model in human readable form
+-- | Show a model in human readable form. Ignore bindings to those variables that start
+-- with "__internal_sbv_"; as these are only for internal purposes
 showModel :: SMTConfig -> SMTModel -> String
-showModel cfg m = intercalate "\n" (map shM assocs ++ concatMap shUI uninterps ++ concatMap shUA arrs)
-  where assocs     = modelAssocs m
-        uninterps  = modelUninterps m
-        arrs       = modelArrays m
-        shM (s, v) = "  " ++ s ++ " = " ++ shCW cfg v
+showModel cfg = intercalate "\n" . display . map shM . filter (not . ignore) . modelAssocs
+  where ignore (s, _) = "__internal_sbv_" `isPrefixOf` s
+        shM (s, v)    = let vs = shCW cfg v in ((length s, s), (vlength vs, vs))
+        display svs   = map line svs
+           where line ((_, s), (_, v)) = "  " ++ right (nameWidth - length s) s ++ " = " ++ left (valWidth - lTrimRight (valPart v)) v
+                 nameWidth             = maximum $ 0 : [l | ((l, _), _) <- svs]
+                 valWidth              = maximum $ 0 : [l | (_, (l, _)) <- svs]
+        right p s = s ++ replicate p ' '
+        left  p s = replicate p ' ' ++ s
+        vlength s = case dropWhile (/= ':') (reverse s) of
+                      (':':':':r) -> length (dropWhile isSpace r)
+                      _           -> length s -- conservative
+        valPart ""          = ""
+        valPart (':':':':_) = ""
+        valPart (x:xs)      = x : valPart xs
+        lTrimRight = length . dropWhile isSpace . reverse
 
 -- | Show a constant value, in the user-specified base
 shCW :: SMTConfig -> CW -> String
@@ -389,7 +398,7 @@ pipeProcess cfg execName opts script cleanErrs = do
                         Left s                          -> return $ Left s
                         Right (ec, contents, allErrors) ->
                           let errors = dropWhile isSpace (cleanErrs allErrors)
-                          in case (null errors, xformExitCode (solver cfg) ec) of
+                          in case (null errors, ec) of
                                 (True, ExitSuccess)  -> return $ Right $ map clean (filter (not . null) (lines contents))
                                 (_, ec')             -> let errors' = if null errors
                                                                       then (if null (dropWhile isSpace contents)
@@ -412,6 +421,47 @@ pipeProcess cfg execName opts script cleanErrs = do
   where clean = reverse . dropWhile isSpace . reverse . dropWhile isSpace
         line  = replicate 78 '='
 
+-- | The standard-model that most SMT solvers should happily work with
+standardModel :: (Bool -> [(Quantifier, NamedSymVar)] -> [String] -> SMTModel, SW -> String -> [String])
+standardModel = (standardModelExtractor, standardValueExtractor)
+
+-- | Some solvers (Z3) require multiple calls for certain value extractions; as in multi-precision reals. Deal with that here
+standardValueExtractor :: SW -> String -> [String]
+standardValueExtractor _ l = [l]
+
+-- | A standard post-processor: Reading the lines of solver output and turning it into a model:
+standardModelExtractor :: Bool -> [(Quantifier, NamedSymVar)] -> [String] -> SMTModel
+standardModelExtractor isSat qinps solverLines = SMTModel { modelAssocs = map snd $ sortByNodeId $ concatMap (interpretSolverModelLine inps) solverLines }
+         where sortByNodeId :: [(Int, a)] -> [(Int, a)]
+               sortByNodeId = sortBy (compare `on` fst)
+               inps -- for "sat", display the prefix existentials. For completeness, we will drop
+                    -- only the trailing foralls. Exception: Don't drop anything if it's all a sequence of foralls
+                    | isSat = map snd $ if all (== ALL) (map fst qinps)
+                                        then qinps
+                                        else reverse $ dropWhile ((== ALL) . fst) $ reverse qinps
+                    -- for "proof", just display the prefix universals
+                    | True  = map snd $ takeWhile ((== ALL) . fst) qinps
+
+-- | A standard engine interface. Most solvers follow-suit here in how we "chat" to them..
+standardEngine :: String
+               -> String
+               -> ([String] -> Int -> [String])
+               -> (Bool -> [(Quantifier, NamedSymVar)] -> [String] -> SMTModel, SW -> String -> [String])
+               -> SMTEngine
+standardEngine envName envOptName addTimeOut (extractMap, extractValue) cfg isSat qinps skolemMap pgm = do
+    execName <-                    getEnv envName     `C.catch` (\(_ :: C.SomeException) -> return (executable (solver cfg)))
+    execOpts <- (splitArgs `fmap`  getEnv envOptName) `C.catch` (\(_ :: C.SomeException) -> return (options (solver cfg)))
+    let cfg'    = cfg {solver = (solver cfg) {executable = execName, options = maybe execOpts (addTimeOut execOpts) (timeOut cfg)}}
+        tweaks  = case solverTweaks cfg' of
+                    [] -> ""
+                    ts -> unlines $ "; --- user given solver tweaks ---" : ts ++ ["; --- end of user given tweaks ---"]
+        cont rm = intercalate "\n" $ concatMap extract skolemMap
+           where extract (Left s)        = extractValue s $ "(echo \"((" ++ show s ++ " " ++ mkSkolemZero rm (kindOf s) ++ "))\")"
+                 extract (Right (s, [])) = extractValue s $ "(get-value (" ++ show s ++ "))"
+                 extract (Right (s, ss)) = extractValue s $ "(get-value (" ++ show s ++ concat [' ' : mkSkolemZero rm (kindOf a) | a <- ss] ++ "))"
+        script = SMTScript {scriptBody = tweaks ++ pgm, scriptModel = Just (cont (roundingMode cfg))}
+    standardSolver cfg' script id (ProofError cfg') (interpretSolverOutput cfg' (extractMap isSat qinps))
+
 -- | A standard solver interface. If the solver is SMT-Lib compliant, then this function should suffice in
 -- communicating with it.
 standardSolver :: SMTConfig -> SMTScript -> (String -> String) -> ([String] -> a) -> ([String] -> a) -> IO a
